/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <map>

// -------------------------------------------------
// Intermediate Edge structure for hashed adjacency
// -------------------------------------------------

/**
 * @brief The Edge struct
 * has to vertices, the first one is always the smaller one
 */
struct Edge {
public:
    /**
   * @brief Edge
   * @param v0 vertex 1
   * @param v1 vertex 2
   * @param _size edge size
   */
  inline Edge (unsigned int v0, unsigned int v1, float _size = 1.) {
    if (v0 < v1) {v[0] = v0; v[1] = v1; } else {v[0] = v1; v[1] = v0; }
    size = _size;
  }

  /**
   * @brief Edge
   * @param e
   */
  inline Edge (const Edge & e) { v[0] = e.v[0]; v[1] = e.v[1]; size = e.size; }
  inline virtual ~Edge () {}

  /**
   * @brief operator =
   * @param e
   * @return this = e;
   */
  inline Edge & operator= (const Edge & e) { v[0] = e.v[0]; v[1] = e.v[1]; size = e.size; return (*this); }

  /**
   * @brief operator ==
   * @param e
   * @return true if (e.v0 = this.v0 and e.v1 = this.v1)
   */
  inline bool operator== (const Edge & e) { return (v[0] == e.v[0] && v[1] == e.v[1]); }

  /**
   * @brief operator <
   * @param e
   * @return true if this smaller than e.
   */
  inline bool operator< (const Edge & e) const { return (v[0] < e.v[0] || (v[0] == e.v[0] && v[1] < e.v[1])); }

  /**
   * @brief contains
   * @param i
   * @return true if the edge contains the vertex i
   */
  inline bool contains (unsigned int i) const { return (v[0] == i || v[1] == i); }

  unsigned int v[2]; /**< vertices of the edge*/
  float size;  /**< edge length*/
};

struct compareEdge {
  inline bool operator()(const Edge e1, const Edge e2) const {
    if (e1.v[0] < e2.v[0])
      return true;
    if (e1.v[0] > e2.v[0])
      return false;
    if (e1.v[1] > e2.v[1])
      return true;
    return false;
  }
};

/**
 * @brief EdgeMapIndex
 */
typedef std::map<Edge, unsigned int, compareEdge> EdgeMapIndex;

/**
 * @brief EdgeWeights
 */
typedef std::map<Edge, float, compareEdge> EdgeWeights;

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
