/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PATCHSELECTION_H
#define PATCHSELECTION_H
#include "Selection.h"
#include <queue>
#include <map>
#include "Histogram2D.h"
#include "SVMclassification.h"
#include <QProgressDialog>
#if WINDOWS_USED
#include "ICP_windows/icpPointToPlane.h"
#else
#include "ICP/icpPointToPlane.h"
#endif

/**
 * @brief The PATCH_SIM_ALGO enum  :
 *
 * BASIC_ALGO : the simplest algorithm (just put a seed on center of the similar patch
 * and run the patch selection algo
 * --> work only for very basic patches
 *
 * SVM : perform an SVM to retrieve the patch boundaries
 *
 * EXPORT_SVM : used to write the svm data in a file
 * (useful if you want to try them in the programs given with libsvm_version)
 *
 * ICP : perform the ICP to retrieve the patch boundaries
 * after the icp we simply transform the user strokes in the local frame
 * and use them to run the selection algorithm on the patch
 *
 * ICP_SVM : perform an ICP and then an SVM to retrieve the exact boundaries
 *
 */
enum PATCH_SIM_ALGO{
    BASIC_ALGO = 0,
    SVM = 1,
    EXPORT_SVM = 2,
    ICP = 3,
    ICP_SVM = 4
};

/**
 * @brief The PatchSelection class
 */
class PatchSelection : public Selection
{
public:

    /**
     * @brief The timing_sim struct
     * sel_nb: number of selections performed by the user
     * sim_nb : number of similar selections
     */
    struct timing_sim{
        unsigned int sel_nb;
        unsigned int sim_nb;
        float time;
    };

    /**
     * @brief PatchSelection
     */
    PatchSelection():Selection(){
        _similarityAlgo = ICP_SVM;
    }

    ~PatchSelection(){}

    /**
     * @brief init initialize all the needed parameters for the current mesh
     * @param m current mesh
     */
    void init(const Mesh &m);

    /**
     * @brief clear
     */
    void clear();

    /**
     * @brief saveTimings
     * @param filename
     */
    void saveTimings(const QString &filename);

    //***************************************************
    //                  SELECTION
    //***************************************************
    /**
     * @brief startSelection start the patch selection
     * @param stroke user stroke used to retrieve the selection
     */
    void startSelection(UserStroke & stroke);

    /**
     * @brief updateSelection update the patch selection
     * @param stroke user stroke used to retrieve the selection
     */
    void updateSelection(UserStroke & stroke);

    /**
     * @brief refineSelection refine the boundaries of the selection
     * @param stroke same as update selection for patches. TODO check
     */
    void refineSelection(UserStroke & stroke);

    /**
     * @brief endSelection end the selection
     * @param stroke
     */
    void endSelection(UserStroke &stroke);

    /**
     * @brief eraseSelection deselect the triangles under the stroke
     * @param stroke
     */
    void eraseSelection(UserStroke & stroke);

    /**
     * @brief clearSelections
     */
    void clearSelections();


    //***************************************************
    //                  EXPANSION
    //***************************************************
    /**
     * @brief findSimilarSelections find the similar patches to the reference
     * @param stroke
     */
    void findSimilarSelections(UserStroke& stroke);

    /**
     * @brief changeSimiliarityThreshold
     * @param val new number of selections to find
     */
    void changeSimiliarityThreshold(int val);

    /**
     * @brief setSimilarityAlgo change the algorithm to find similar patches
     * @param algo algorithm to use
     * @param stroke stroke used to retrieve the new similar selections
     */
    void setSimilarityAlgo(PATCH_SIM_ALGO algo, UserStroke& stroke);

    /**
     * @brief getPatchSimAlgo
     * @return the current type of the algorithm used to find similar patches
     */
    PATCH_SIM_ALGO getPatchSimAlgo(){return _similarityAlgo;}

    /**
     * @brief getSimilarityDist
     * @return the distance between the reference descriptor and the
     * descriptor of the vertex currently viewed (under the mouse for example)
     */
    const std::vector<float>& getSimilarityDist(){return similarity_dist;}

    /**
     * @brief getRefHistogram
     * @return the reference histogram
     */
    QPixmap getRefHistogram() {
        if (refVertex < vertices.size())
            return _shapeContexts[refVertex].drawHistogram();
        else
            return QPixmap();
    }

    /**
     * @brief getUnderMouseHistogram return the histogram of the vertex under the mouse
     * @param v vertex under the mouse
     * @param dist distance between the current histogram and the reference one
     * @return the current histogram
     */
    QPixmap getUnderMouseHistogram(unsigned int v, float & dist){
        if (!_sc_computed[v])
            computeSC(v);
        dist = similarity_dist[v];
        return _shapeContexts[v].drawHistogram();
    }

    /**
     * @brief openSVMselection
     * @param filename
     */
    void openSVMselection(const QString & filename);

    /**
     * @brief saveSelectionHistImage
     * save the histogram in an image
     */
    void saveSelectionHistImage();

    /**
     * @brief drawICPDebug
     * draw visual hints to understand the result of the ICP
     */
    void drawICPDebug();

    /**
     * @brief drawShapeContextDebug
     * for now only draw a sphere on the current potiental center
     */
    void drawShapeContextDebug();


protected:
    //***************************************************
    //                  SELECTION
    //***************************************************

    // faces are marked as follow:
    /**
     * @brief initializeElementsToClassify
     * @param starting_point the last stroke point (ind) that was previously treated
     * @param points the user stroke points
     * @param viewedFaces : faces as marked as follow:
     *      -1 : not viewed
     *       0 : background
     *       1 : unclassied
     *       2 : foreground
     *       3 : currently being classified as B, F, U
     * @param unsortedFaces
     */
    void initializeElementsToClassify( unsigned int starting_point,
                                       std::vector< UserStroke::strokePoints>& points,
                                       std::vector< int > & viewedFaces ,
                                      std::vector< unsigned int > & unsortedFaces);


    /**
     * @brief findTrianglesInsideSphere find all the triangles t under the sphere
     * that are not already part of the selection (ie. inside[t]!=inside_value)
     * and set them to inside[t] = new_inside_value
     * @param point sphere center
     * @param radius sphere radius
     * @param inside vector of viewedFaces
     * @param inside_value previous inside value ( never put 4 )
     * @param new_inside_value new inside value ( never put 4 )
     * @return true if there is at least one triangle
     * under the sphere that is not already part of the selection
     */
    bool findTrianglesInsideSphere( Vec3Df point, float radius,
                                    std::vector<int> &  inside,
                                    int inside_value,
                                    int new_inside_value);

    /**
     * @brief classifyUnsortedFaces classify the unsorted faces as selected or not
     * @param faces same as viewedFaces
     * @param unsortedFaces
     */
    void classifyUnsortedFaces(std::vector< int > & faces,
                               std::vector< unsigned int> const & unsortedFaces );

    void findVerticesInsideDisk( std::vector<unsigned int> &vertices_ind, unsigned int v_indice );


    //***************************************************
    //                  EXPANSION
    //***************************************************

    /**
     * @brief recomputeSimilarityDist
     */
    void recomputeSimilarityDist();

    /**
     * @brief setReference
     * set the refVertex, refCenter and refRadius
     */
    void setReference();

    /**
     * @brief selectKiemSimilarPatch
     * @param sortedSim vector of sorted pairs
     *  each pair contains the index of a vertex and it distance to the PSC ref
     * @param vertices_ind_ref
     * @return the distance between the similar patch found and the ref one
     */
    float selectKiemSimilarPatch( const std::vector<std::pair<float, unsigned int> > & sortedSim,
                                  const std::vector<unsigned int> & vertices_ind_ref);

    /**
     * @brief computeICP_SVM find the boundaries of the k° patch
     * using the ICP + SVM algo
     * @param v_indice index of the center of the similar patch
     * @param vertices_ind_ref vertices of the set of the reference patch
     */
    void computeICP_SVM(unsigned int v_indice,
                        const std::vector<unsigned int> & vertices_ind_ref);

    /**
     * @brief basicSimilarityAlgo
     * @param nnv
     * @param negSeeds
     * @param center
     */
    void basicSimilarityAlgo( std::vector<unsigned int> nnv,
                              std::vector<unsigned int> negSeeds,
                              unsigned int center);

    //***************************************************
    //              PATCH SHAPE CONTEXT
    //***************************************************
    /**
     * @brief computeShapeContexts compute the PSC for each vertex of the shape
     */
    void computeShapeContexts();

    /**
     * @brief computeSCangle compute the angle between the normal of the vertex v and the
     * one of the reference vertex
     * @param v index of the current vertex
     * @param n_ref normal of the reference vertex
     * @return
     */
    float computeSCangle(unsigned int v, Vec3Df n_ref );

    /**
     * @brief computeSC compute the PSC of vertex of v
     * @param v index of the vertex
     */
    void computeSC(unsigned int v);

    //***************************************************
    //                  SVM STUFFS
    //***************************************************

    /**
     * @brief writeSVM write the labels and vectors of the vertices inside the
     * set of vertices of the center v_indice.
     * useful to export the values and visualize them for example
     * @param fileName name of the file
     * @param v_indice index of the center vertex
     */
    void writeSVM(const QString &  fileName, unsigned int v_indice);

    /**
     * @brief trainSVM train the svm
     */
    void trainSVM();

    /**
      * @brief predictSVM SVM algorithm prediction (without the ICP)
      * @param v_ind index of the center of the current patch
      */
     void predictSVM(unsigned int v_ind);

    /**
     * @brief getValuesforSVM_PCAFrame compute the vector in the feature space
     * for each vertex in the set
     * @param v_indice index of the center vertex
     * @param _labels 1 if a vertex is selected 0 otherwise (only useful for the ref)
     * @param _vectors positions of the vertices of the set in the feature space
     * @param vertices_ind indices of all the vertices found inside the set
     */
    void getValuesforSVM_PCAFrame(unsigned int v_indice,
                                  std::vector< int > &_labels,
                                  std::vector<std::vector<double> > &_vectors,
                                  std::vector<unsigned int> &vertices_ind);


    /**
     * @brief getValuesforSVM_withICP
     * @param _labels 1 if a vertex is selected 0 otherwise (only useful for the ref)
     * @param _vectors positions of the vertices of the set in the feature space
     * @param vertices_ind indices of all the vertices found inside the set
     * @param eVec frame that was previously computed by the registration step (PCA+ICP)
     * @param center center that was previously computed by the registration step
     */
    void getValuesforSVM_withICP(std::vector< int > &_labels,
                                 std::vector<std::vector<double> > &_vectors,
                                 std::vector<unsigned int> &vertices_ind,
                                 std::vector<Vec3Df> eVec,
                                 Vec3Df center);

    /**
     * @brief initICP_SVM train the svm with the reference points
     * @param vertices_ind (in) set of vertices in the set
     * @param frame (in) frame of the set
     * @param center (in) center of the set
     */
    void initICP_SVM(std::vector<unsigned int> & vertices_ind,
                     std::vector<Vec3Df> & frame,
                     Vec3Df & center);


    //***************************************************
    //                  ICP STUFFS
    //***************************************************

    /**
     * @brief computeWeightAndNormals
     * find the normals and the normalized weights associated to the vertices
     *  in the v_indices set
     * @param v_indices set of vertices
     * @param normals (out) normals (smoothed) of the vertices
     * @param weights (out) weights (normalized) of the vertices
     */
    void computeWeightAndNormals(const std::vector<unsigned int >& v_indices,
                                 std::vector<Vec3Df> & normals,
                                 std::vector<double> &weights);

    /**
     * @brief findLocalFrame find the frame of the current patch
     * @param vertices_ind (out) indices of the vertices in the set
     * @param v_indice index (in) of the center of the current patch
     * @param frame (out) computed frame
     * @param flipAxis  = 0 if no flip
     *                  = 1 if eVec[1] flipped
     *                  = 2 if eVec[2] flipped
     */
    void findLocalFrame( std::vector<unsigned int> & vertices_ind, unsigned int v_indice,
                         std::vector<Vec3Df> & frame, unsigned int & flipAxis);

    /**
     * @brief PatchSelection::initICPref compute the set of vertices,
     * it frame and it center
     * @param vertices_ind (out) vertices of the reference set
     * @param frame (out) frame of the reference set
     * @param center (out) center of the reference set
     */
    void initICPref(std::vector<unsigned int> & vertices_ind,
                    std::vector<Vec3Df> & frame,
                    Vec3Df & center);

    /**
     * @brief computeICP compute the result of the ICP on the reference frame
     * @param v_indice (in) index of the center of the current patch
     * @param vertices_ind_ref (in) indices of the vertices of the ref set
     * @param vertices_ind (in) indices of the vertices of the local set
     * @param firstTransform (out) first transformation (result of the PCA)
     * @param icpTransform (out) transformation resulting of the ICP
     * @param loc_frame_icp (out) local frame after the ICP
     * @param centerICP (out) local center after the ICP
     */
    void computeICP(unsigned int v_indice,
                    const std::vector<unsigned int> & vertices_ind_ref,
                    std::vector<unsigned int> & vertices_ind,
                    Matrix & firstTransform,
                    Matrix & icpTransform,
                    std::vector<Vec3Df> & loc_frame_icp,
                    Vec3Df & centerICP );

    /**
     * @brief computeBasicAlgoFromICP transform the user strokes
     * so that they are on the local frame
     * then use these strokes to run the selection algorithm and
     * retrieve the local patch
     * @param firstTransform transformation before ICP
     * @param icpTransform transformation after ICP
     */
    void computeBasicAlgoFromICP(Matrix & firstTransform,
                                 Matrix & icpTransform );



private:
    // for the selection
    unsigned int strokeConnectCompIndice;  /**< indice of the connected component under the user stroke*/
    float angle;  /**<  for the L2_1 or L_cross distance*/
    float diskRadius;  /**<  radius of the selection *1.2 */
    bool LcrossDistUsed;  /**< true if te L_cross dist is used */

    std::vector<float> trianglesUnderMouse; /**< triangles under the user mouse*/
    std::vector<int> verticeInSelection; /**< >0 if a vertex is in the selection */

    // for the expansion
    Vec3Df refCenter;  /**< center of the reference patch*/
    unsigned int refVertex;  /**< index of the vertex at the center of the ref patch */
    unsigned int ind_vertex;/**< index of the vertex at the center of the local patch*/

    PATCH_SIM_ALGO _similarityAlgo; /**< algorithm used to find similar patches*/
    std::vector<float> similarity_dist; /**< distance between each vertex and the ref one*/

    Vec3Df PCAeigenVec[3]; /**< eigen vectors, resulting of the PCA on the local patch*/
    Vec3Df PCAeigenVecRef[3]; /**< eigen vectors, resulting of the PCA on the ref patch*/

    // PSC
    std::vector<Hist2Df> _shapeContexts;  /**< PSC of all the vertices*/
    std::vector<bool> _sc_computed; /**< true if the PSC of vertex i is compued*/
    Hist2Df refSC; /**< PSC of the reference patch*/

    // svm
    SVMClassification _svm; /**< svm structure*/
    bool use_PK; /**< 1 if we use a precomputed kernel for the SVM*/

    // icp stuff
    float momentsRef[4]; /**< moments of the reference */
    std::vector<Vec3Df> ICPpointsRef; /**< points of the ref set before icp*/
    std::vector<Vec3Df> ICPpointsRefTransformed; /**<  points of the ref set after icp*/
    Vec3Df PCAeigenVecDeformed[3]; /**< position of the local frame after icp */
    Vec3Df centerVecDeformed; /**< position of the local center after icp*/

    std::vector<unsigned int> NegSeedsToDraw; /**< for debug*/
    std::vector< UserStroke::strokePoints > strokeForSimiliarity; /**< strokes for the ICP only algo*/

    // timing
    std::vector< timing_sim > timing_SVM_train; /**< svm train timing */
    std::vector< timing_sim > timing_SVM_predict; /**<  svm predict timing*/
    std::vector< timing_sim > timing_ICP; /**< icp timing */
    std::vector< std::pair< float, timing_sim> > timing_SC; /**< PSC timing*/
};

#endif // PATCHSELECTION_H
