/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SELECTIONSMANAGER_H
#define SELECTIONSMANAGER_H
#include "SelectionWidget.h"
#include "PatchSelection.h"
#include "PartSelection.h"
#include "CompSelection.h"

/**
 * @brief The TYPE enum
 * type of selection
 */
enum TYPE{
    COMP = 0,
    PATCH = 1,
    PART = 2,
    NONE = 3
};

/**
 * @class SelectionManager
 * @brief manage the current selection on the model
 */
class SelectionsManager : public QObject
{
    Q_OBJECT
public:

    SelectionsManager(){
        _type = PATCH;
        _comp = new CompSelection;
        _patch = new PatchSelection;
        _part = new PartSelection;
    }

    /**
     * @brief init
     * @param m
     */
    void init (const Mesh & m);

    /**
     * @brief saveTimings
     * @param filename
     */
    void saveTimings(const QString &filename){
        _patch->saveTimings(filename);
        _part->saveTimings(filename);
        _comp->saveTimings(filename);
    }
    //***************************************************
    //                  SELECTION ACCESSORS
    //***************************************************

    /**
     * @fn getSelections
     * @return the selections on the model
     */
    const std::vector<float> & getSelections(){return _selections;}

    /**
     * @fn setSelections
     * @param selection the selections to apply on the model
     */
    void setSelections(const std::vector<float> & selections){
        _selections.clear();
        for (unsigned int i = 0; i < selections.size(); i++)
            _selections.push_back(selections[i]);
    }

    /**
     * @fn getCurrentSelections
     * @return the current selection on the model
     */
    const std::vector<float> & getCurrentSelections(){return _currentSelections;}

    /**
     * @brief getCurrentSelectionType
     * @return the type of selection currently used
     */
    TYPE getCurrentSelectionType(){return _type;}

    /**
     * @brief setSelectionType set the selection type to t
     * @param t new selection type
     */
    void setSelectionType(TYPE t){_type = t;}

    /**
     * @brief selectionIsEmpty
     * @return true if there is no current selection
     */
    bool selectionIsEmpty(){
        bool result = true;
        for (unsigned int i = 0; i< _currentSelections.size(); i++ )
            if (_currentSelections[i]!=-1)
                result = false;
        return result;
    }

    //***************************************************
    //                  FIELDS ACCESSORS
    //***************************************************

    /**
     * @fn getUserField
     * @return the field defined by the user stroke (works only if the selection is a part)
     */
    const std::vector<float> & getUserField()const {return _part->getUserField();}

    /**
     * @fn getConformalFactor
     * @return the model conformal factor (computed like in "characterizing shape using conformal factor")
     */
    const std::vector<float> & getConformalFactor()const {return _part->getConformalFactor();}

    /**
     * @brief getPatchSimField
     * @return the scalar field representing the similarity for patches
     */
    const std::vector<float> & getPatchSimField() const {return _patch->getSimilarityDist();}

    /**
     * @brief getCompSimField
     * @return the scalar field representing the similarity for components
     */
    const std::vector<float> & getCompSimField() const {return _comp->getSimilarityDist();}


    //***************************************************
    //                  HISTOGRAMS ACCESSORS
    //***************************************************

    /**
     * @fn getRefHistogram
     * @return the descriptor (2D histogram) of the reference vertex (at the center of the user defined selection)
     */
    QPixmap getRefHistogram(){return _curSelection->getRefHistogram();}

    /**
     * @fn getUnderMouseHistogram
     * @return the descriptor (2D histogram) of the vertex under the user mouse
     */
    QPixmap getUnderMouseHistogram(unsigned int v, float & dist){
        return _curSelection->getUnderMouseHistogram(v, dist);}

    /**
     * @brief changeHistSimDist change the distance to mesure histogram differences
     * @param val new distance to use
     */
    void changeHistSimDist(int val){
        _curSelection->changeHistSimDist(val);
    }

    //***************************************************
    //                  SELECTION
    //***************************************************
    /**
     * @brief startSelection start the current selection
     * @param _widget
     */
    void startSelection(SelectionWidget* _widget);

    /**
     * @brief updateSelection
     * when the user moves the mouse
     * @param _widget
     */
    void updateSelection(SelectionWidget* _widget);

    /**
     * @brief endSelection stop the current selection
     * when the user releases the mouse
     * @param _widget
     */
    void endSelection(SelectionWidget* _widget);

    /**
     * @brief clearCurrentSelections
     */
    void clearCurrentSelections();

    /**
     * @brief saveCurrentSelections
     */
    void saveCurrentSelections();


    /**
     * @fn reverseCurrentSelection
     * @brief reverse the current selection
     */
    void reverseCurrentSelection();

    /**
     * @brief deleteAll
     */
    void deleteAll();
    //***************************************************
    //                  EXPANSION
    //***************************************************

    /**
     * @brief searchSimilarSelections
     * @param stroke
     */
    void searchSimilarSelections(UserStroke & stroke);

    /**
     * @brief changeSimilarityThreshold
     * @param val number of similar selections asked
     */
    void changeSimilarityThreshold(int val);

    /**
     * @brief changePatchSimAlgo
     * @param i
     * @param stroke
     */
    void changePatchSimAlgo( PATCH_SIM_ALGO i, UserStroke& stroke){
        _patch->setSimilarityAlgo(i, stroke);
        _currentSelections = _curSelection->getSelections();
    }

    /**
     * @brief openSVMselection
     * @param filename
     */
    void openSVMselection(const QString & filename);

    /**
     * @brief saveSelectionHist
     */
    void saveSelectionHist(){
        if (_type == PATCH)
            _patch->saveSelectionHistImage();
    }

    /**
     * @brief getPatchSimAlgo
     * @return the similarity algorithm currently used for patches
     */
    PATCH_SIM_ALGO getPatchSimAlgo(){ return _patch->getPatchSimAlgo();}

    //***************************************************
    //                  DRAWING FUNCTIONS
    //***************************************************
    /**
     * @brief drawPartUserConstraints draw the user constraints for parts
     */
    void drawPartUserConstraints(){_part->drawUserConstraints();}

    /**
     * @brief drawPartIsolines draw the isoline from the user field for parts
     */
    void drawPartIsolines(){_part->drawIsolines();}

    /**
     * @brief drawPartSimilarityConstraints raw the similarity constraints for parts
     */
    void drawPartSimilarityConstraints(){_part->drawSimilarConstraints();}

    /**
     * @brief drawPatchSCDebug draw the patch shape context
     */
    void drawPatchSCDebug(){_patch->drawShapeContextDebug();}

    /**
     * @brief drawPatchICPDebug
     */
    void drawPatchICPDebug(){ _patch->drawICPDebug(); }

    /**
     * @brief drawCompSCDebug
     */
    void drawCompSCDebug(){_comp->drawShapeContext(0);}

protected:
    /**
     * @brief recordCurrentSelections save selections
     */
    void recordCurrentSelections();

    /**
     * @brief deleteSelectionUnderStroke
     * @param stroke
     */
    void deleteSelectionUnderStroke(UserStroke& stroke);

private:
  std::vector<float> _selections; /**< face indices : all the selections */
  std::vector<float> _currentSelections; /**< face indices : current selection */
  Selection* _curSelection; /**<  current selection can be either _patch, _part, _comp*/
  PatchSelection* _patch; /**< patch selection type */
  PartSelection* _part; /**< part selection type */
  CompSelection* _comp; /**< connected component selection type */
  unsigned int nnsSelectionsNb; /**< number of selections made by the user */
  unsigned int allSelectionsNb; /**< number of all selections (automatically detected and hand drawn) */
  TYPE _type; /**< the type of the current selection */
  Mesh mesh; /**< the mesh */

};

#endif // SELECTIONSMANAGER_H
