/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "Shader.h"
#include "Debug.h"
#include "OpenGLHeader.h"

#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <string.h>
#include <fcntl.h>

#include <sys/types.h>
#include <unistd.h>

#ifdef _WIN32 /*[*/
#include <io.h>
#endif /*]*/

using namespace std;

Shader::Shader () : shaderProgram (0), vertexShader (0), fragmentShader (0),
                    vertexShaderSize (0), fragmentShaderSize (0) {}


Shader::~Shader () {
  if (hasVertexShader ())
    glDeleteShader (vertexShader);
  if (hasFragmentShader ())
    glDeleteShader (fragmentShaderSize);
  glDeleteProgram (shaderProgram);
}


void Shader::loadFromFile (const string & vertexShaderFilename, const string & fragmentShaderFilename) {
  const GLchar * vertexShaderSource;
  const GLchar * fragmentShaderSource;
  if (vertexShaderFilename != "")
    vertexShaderSource = readShaderSource (vertexShaderFilename, vertexShaderSize);
  if (fragmentShaderFilename != "")
    fragmentShaderSource = readShaderSource (fragmentShaderFilename, fragmentShaderSize);

  shaderProgram = glCreateProgram ();
  if (hasVertexShader () == true) {
    compileAttach (vertexShader, GL_VERTEX_SHADER, &vertexShaderSource);
    delete [] vertexShaderSource;
  }
  if (hasFragmentShader () == true) {
    compileAttach (fragmentShader, GL_FRAGMENT_SHADER, &fragmentShaderSource);
    delete [] fragmentShaderSource;
  }
  glLinkProgram (shaderProgram);
   GLint iLinked;
  glGetObjectParameterivARB( shaderProgram, GL_OBJECT_LINK_STATUS_ARB, &iLinked );

  GLint iValidate;
  glGetObjectParameterivARB( shaderProgram, GL_VALIDATE_STATUS, &iValidate );

  // This can have warnings and errors!
  if ( !iLinked ) {
      CheckError();
      std::cout << " error when linking shader \n";
  }

  GetOpenGLError();
}


void Shader::bind () {
  GetOpenGLError();
  glUseProgram (shaderProgram);
}


void Shader::unbind () {
  glUseProgram (0);
}


// ------------------
// Protected methods.
// ------------------

unsigned int Shader::getShaderSize (const string & filename) {
  int fd;
  unsigned int count = 0;
#ifdef _WIN32
  fd = _open (filename.c_str (), _O_RDONLY);
  if (fd != -1) {
    count = static_cast<unsigned int>(_lseek (fd, 0, SEEK_END) + 1);
    _close(fd);
  } else 
    throw ShaderException (string ("getShaderSize: bad Shader File Name") + filename); 
#else 
  fd = open (filename.c_str (), O_RDONLY);
  if (fd != -1) {
    count = static_cast<unsigned int>(lseek (fd, 0, SEEK_END) + 1);
    close(fd);
  } else 
    throw ShaderException (string ("getShaderSize: bad Shader File Name") + filename);
#endif
  return count;
}


GLchar * Shader::readShaderSource(const string & shaderFilename, unsigned int & shaderSize) {
  shaderSize = getShaderSize (shaderFilename);
  FILE * fh = fopen (shaderFilename.c_str (), "r");
  if (!fh)
    throw ShaderException (string ("readShaderSource : Bad Shader Filename ") + shaderFilename);
    
  GLchar * shaderSource = new GLchar[shaderSize];
  fseek (fh, 0, SEEK_SET);
  int count = fread (shaderSource, 1, shaderSize, fh);
  shaderSource[count] = '\0';
  fclose (fh);
  //  if (ferror (fh))
  //  throw ShaderException (string ("readShaderSource : error while reading file") + shaderFilename);

  return shaderSource;
}


GLint Shader::getUniLoc (GLuint program, const GLchar *name){
  GLint loc = glGetUniformLocation (program, name);
  if (loc == -1)
      Trace(1, " No such uniform named " << name );

  GetOpenGLError();
 // printOpenGLError();
  return loc;
}

void Shader::bindFragDataLocation(GLuint program, GLuint number, const GLchar* name){
    glBindFragDataLocation(program, number, name);

    GetOpenGLError();
   // printOpenGLError();
}

GLint Shader::getFragDataLocation(GLuint program, const char *name){
    GLint loc =  glGetFragDataLocation(program, name);
//    if (loc == -1)
//      throw ShaderException (string ("No such uniform named") + string (name));

    GetOpenGLError();
   // printOpenGLError();
    return loc;
}

void Shader::printShaderInfoLog (GLuint shader) {
  int infologLength = 0;
  int charsWritten  = 0;
  GLchar *infoLog;
  

  GetOpenGLError();
  //printOpenGLError ();  // Check for OpenGL errors
  glGetShaderiv (shader, GL_INFO_LOG_LENGTH, &infologLength);

  GetOpenGLError();
  //printOpenGLError ();  // Check for OpenGL errors

  if (infologLength > 0) {
    infoLog = new GLchar[infologLength];
    if (infoLog == NULL) {
      printf("ERROR: Could not allocate InfoLog buffer\n");
      exit(1);
    }
    glGetShaderInfoLog (shader, infologLength, &charsWritten, infoLog);
    //    cerr << "InfoLog:" << endl << infoLog << endl << endl;
    delete [] infoLog;
  }

  GetOpenGLError();
 // printOpenGLError();  // Check for OpenGL errors
}

void Shader::printProgramInfoLog (GLuint program) {
  int infologLength = 0;
  int charsWritten  = 0;
  GLchar *infoLog;
  

  GetOpenGLError();
  //printOpenGLError ();  // Check for OpenGL errors
  glGetProgramiv (program, GL_INFO_LOG_LENGTH, &infologLength);

  GetOpenGLError();
  //printOpenGLError ();  // Check for OpenGL errors

  if (infologLength > 0) {
    infoLog = new GLchar[infologLength];
    if (infoLog == NULL) {
      printf("ERROR: Could not allocate InfoLog buffer\n");
      exit(1);
    }
    glGetProgramInfoLog (program, infologLength, &charsWritten, infoLog);
    //cerr << "InfoLog:" << endl << infoLog << endl << endl;
    delete [] infoLog;
  }

  GetOpenGLError();
// printOpenGLError();  // Check for OpenGL errors
}

// Error Output
bool Shader::CheckError()
{
    int		iInfoLogLength = 0;
    int		iCharsWritten  = 0;
    char*	szInfoLog = 0;

    glGetObjectParameterivARB( shaderProgram, GL_OBJECT_INFO_LOG_LENGTH_ARB, &iInfoLogLength );
    if ( iInfoLogLength > 1 ) {
        szInfoLog = new char[ iInfoLogLength ];
        glGetInfoLogARB( shaderProgram, iInfoLogLength, &iCharsWritten, szInfoLog );
        //m_strError = szInfoLog;
        std::string m_strError = std::string( szInfoLog );
        std::cout<< " Check error " << m_strError<< "\n";
        delete [] szInfoLog;
        return false;
    }
    return true;
}


void Shader::compileAttach (GLuint & shader, GLenum type, const GLchar ** source) {
  GLint shaderCompiled; //  = 0
  shader = glCreateShader (type);
  glShaderSource (shader, 1, source, NULL);
  glCompileShader (shader);

  GetOpenGLError();
  glGetShaderiv( shader, GL_COMPILE_STATUS, &shaderCompiled );
  //glGetObjectParameterivARB( shaderProgram, GL_OBJECT_COMPILE_STATUS_ARB, &shaderCompiled );

  printShaderInfoLog (shader);

  if (!shaderCompiled)
      std::cout << " something wrong  in compile Attach function \n";

  CheckError();
  GetOpenGLError();


//  printOpenGLError ();  // Check for OpenGL errors
//  glGetShaderiv (shader, GL_COMPILE_STATUS, &shaderCompiled);
//  printOpenGLError ();  // Check for OpenGL errors
//  if (!shaderCompiled)
//    throw ShaderException ("Error: shader not compiled");
  glAttachShader (shaderProgram, shader);
  GetOpenGLError();
}
