/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "SelectionWidget.h"

#include <QtGui>

bool SelectionWidget::addPointToStroke (QMouseEvent* const event, qglviewer::Camera* const cam)
{
    if (lastPointAdded == QPoint(-100, -100)) // we add the first point
    {
        bool found;
        qglviewer::Vec point = cam->pointUnderPixel(event->pos(), found);

        if (found)
        {
            previous_p = Vec3Df(point.x, point.y, point.z);
            _stroke.addPoint(previous_p, radius);
            lastPointAdded = event->pos();
            return true;
        } else {
            return false;
        }
    }  else if ( (event->pos()-lastPointAdded).manhattanLength() >= 4 ) {
        // if the mouse is far enough from it previous position we add a point
        bool found;
        qglviewer::Vec point = cam->pointUnderPixel(event->pos(), found);

        if (found)
        {
            Vec3Df p_cur = Vec3Df(point.x, point.y, point.z);
            float dist = (previous_p-p_cur).getLength();
            if (previous_dist != FLT_MIN)
            {
                if ( 8.f*previous_dist < dist )
                    return false;
            }
            _stroke.addPoint(p_cur, radius);
            lastPointAdded = event->pos();
            previous_p = p_cur;
            previous_dist = dist;
            return true;
        } else
            return false;
    } else {
        return true;
    }
}

bool SelectionWidget::addPointToRefineStroke (QMouseEvent* const event, qglviewer::Camera* const cam)
{
    bool found;
    qglviewer::Vec point = cam->pointUnderPixel(event->pos(), found);
    if (found)
    {
        _refineStroke.addPoint( Vec3Df(point.x, point.y, point.z), radius);
        return true;
    }
    return false;
}

bool SelectionWidget::erasePointsToStroke (QMouseEvent* const event, qglviewer::Camera* const cam)
{
    bool found;
    qglviewer::Vec point = cam->pointUnderPixel(event->pos(), found);
    if (found)
    {
        _stroke.deletePoint(Vec3Df(point.x, point.y, point.z), radius );
        return true;
    } else {
        return false;
    }
}

bool SelectionWidget::deleteStrokeUnderPoint(QMouseEvent* const event, qglviewer::Camera* const cam)
{
    bool found;
    qglviewer::Vec point = cam->pointUnderPixel(event->pos(), found);
    if (found)
    {
        _stroke.deletePoint(Vec3Df(point.x, point.y, point.z), radius);
        return true;
    } else {
        return false;
    }
}

void SelectionWidget::mousePressEvent(QMouseEvent* const event, qglviewer::Camera* const cam )
{
    previous_dist = FLT_MIN;
    if (event->button()==Qt::RightButton)
    {
        selectionStarted = true;
        _mode = ERASE;
        emit eraseMode();
        erasePointsToStroke(event, cam);
    }

    if ((event->button()==Qt::LeftButton)  )
    {
        selectionStarted = true;
        switch(_mode)
        {
        case DRAW:
            addPointToStroke(event, cam);
            break;
        case ERASE:
            erasePointsToStroke(event, cam);
            break;
        case DEL:
            deleteStrokeUnderPoint(event,cam);
            break;
        case REFINE:
            addPointToRefineStroke(event, cam);
            break;
        default:
            break;
        }
    }
    _mouse_released = false;
    emit pressed();
}

void SelectionWidget::mouseMoveEvent(QMouseEvent* const event, qglviewer::Camera* const cam )
{

    Trace (4, "SELECT WIDGET: mouse move event " << _mouse_released);
    bool found;
    qglviewer::Vec point = cam->pointUnderPixel(event->pos(), found);
    if (found)
        center = Vec3Df(point.x, point.y, point.z);

    if (! _mouse_released && selectionStarted )
    {
        bool pointIsAdded ;
        switch(_mode)
        {
        case DRAW:
            pointIsAdded = addPointToStroke(event, cam);
            break;
        case ERASE:
            pointIsAdded = erasePointsToStroke(event, cam);
            break;
        case DEL:
            pointIsAdded = false;
            break;
        case REFINE:
            pointIsAdded = addPointToRefineStroke(event, cam);
            break;
        default:
            pointIsAdded = false;
            break;
        }
        if (!pointIsAdded )
        {
            _mouse_released = true;
            if (_mode == DRAW || _mode == ERASE || _mode == DEL)
                _stroke.lastStrokeDrawn();
            else
                _refineStroke.lastStrokeDrawn();
            emit released();
        } else
            emit moved();
     }
}

void SelectionWidget::mouseReleaseEvent(QMouseEvent* const event, qglviewer::Camera* const cam )
{
    bool found;
    qglviewer::Vec point = cam->pointUnderPixel(event->pos(), found);
    if (found)
        center = Vec3Df(point.x, point.y, point.z);

    if (! _mouse_released && selectionStarted)
    {
        switch(_mode)
        {
        case DRAW:
            addPointToStroke(event, cam);
            break;
        case ERASE:
            erasePointsToStroke(event, cam);
            break;
        case DEL:
            break;
        case REFINE:
            addPointToRefineStroke(event, cam);
        default:
            break;
        }
         if ((event->button()==Qt::RightButton)  )
             _mode = DRAW;
        _mouse_released = true;
        selectionStarted = false;
        if (_mode == DRAW || _mode == ERASE || _mode == DEL)
            _stroke.lastStrokeDrawn();
        else
            _refineStroke.lastStrokeDrawn();
        emit released();
    }
}

void SelectionWidget::wheelEvent(QWheelEvent *const event, qglviewer::Camera *const camera){
    Q_UNUSED(camera);
    if (_state == ACTIVE )
    {
        if (radius + (float) event->delta()/(24000) > 0)
            radius += (float) event->delta()/(24000) ;
    }
    event->ignore();
}
