/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "VisuDock.h"

VisuDock::VisuDock(const QString & title, GLViewer* viewer, QWidget* parent): QDockWidget(title, parent)
{
    _viewer = viewer;
    QWidget * contents = new QWidget();
    // vertical layout
    QVBoxLayout *vbox = new QVBoxLayout(contents);
    // horizontal layout
    QHBoxLayout* hBoxColor = new QHBoxLayout();
    vbox->addLayout(hBoxColor);

    // to change patch color
    QSmartColorPicker* patchColor= new QSmartColorPicker( QColor(222.f, 156.f, 0.f) );
    patchColor->setToolTip("change the color of patches");
    hBoxColor->addWidget(patchColor);
    connect( patchColor , SIGNAL(colorChanged(QColor)) , _viewer , SLOT(updatePatchColor(QColor)) );

    // to change part color
    QSmartColorPicker* partColor= new QSmartColorPicker( QColor( 54.f, 140.f, 23.f) );
    partColor->setToolTip("change the color of the stroke");
    hBoxColor->addWidget(partColor);
    connect( partColor , SIGNAL(colorChanged(QColor)) , _viewer , SLOT(updatePartColor(QColor)) );

    // to change component color
    QSmartColorPicker* compColor= new QSmartColorPicker( QColor(7.f, 113.f, 162.f) );
    compColor->setToolTip("change the color of the stroke");
    hBoxColor->addWidget(compColor);
    connect( compColor , SIGNAL(colorChanged(QColor)) , _viewer , SLOT(updateccColor(QColor)) );

    // to change background color
    QSmartColorPicker* bgColor= new QSmartColorPicker( QColor(83.f, 83.f, 83.f) );
    bgColor->setToolTip("change the color of the stroke");
    hBoxColor->addWidget(bgColor);
    connect( bgColor , SIGNAL(colorChanged(QColor)) , _viewer , SLOT(updatebgColor(QColor)) );

    visuBox = new QButtonGroup();
    QCheckBox* usual_view = new QCheckBox();
    usual_view->setText("selections");
    QCheckBox* harmonic_field = new QCheckBox();
    harmonic_field->setText("show user field");
    QCheckBox* similarity_map = new QCheckBox();
    similarity_map->setText("Similarity map");

    // caution here the id of each button should be the same of the enum number in VISU_MODE (glviewer.h)
    visuBox->addButton(usual_view, 2);
    visuBox->addButton(harmonic_field, 0);
    visuBox->addButton(similarity_map, 1);
    visuBox->setExclusive(true);
    connect(visuBox, SIGNAL(buttonClicked(int)),_viewer, SLOT(changeVisuMode(int)));
    harmonic_field->setChecked(true);
    similarity_map->setChecked(false);

    vbox->addWidget(usual_view);
    vbox->addWidget(harmonic_field);
    vbox->addWidget(similarity_map);

    QFrame* line1 = new QFrame();
    line1->setFrameShape(QFrame::HLine);
    line1->setFrameShadow(QFrame::Sunken);
    vbox->addWidget(line1,1);

    QButtonGroup* patchSimAlgo = new QButtonGroup();
    QCheckBox* basic_algo = new QCheckBox();
    basic_algo->setText("Basic Algo");
    QCheckBox* svm = new QCheckBox();
    svm->setText("SVM");
    QCheckBox* export_svm = new QCheckBox();
    export_svm->setText("Export SVM in file");
    QCheckBox* icp = new QCheckBox();
    icp->setText("ICP");
    QCheckBox* icp_svm = new QCheckBox();
    icp_svm->setText("ICP-SVM");

    patchSimAlgo->addButton(basic_algo, 0);
    patchSimAlgo->addButton(svm,1);
    patchSimAlgo->addButton(export_svm, 2);
    patchSimAlgo->addButton(icp, 3);
    patchSimAlgo->addButton(icp_svm, 4);
    patchSimAlgo->setExclusive(true);
    icp_svm->setChecked(true);
    connect(patchSimAlgo, SIGNAL(buttonClicked(int)),_viewer, SLOT(changePatchSimAlgo(int)));
    vbox->addWidget(basic_algo);
    vbox->addWidget(svm);
    vbox->addWidget(export_svm);
    vbox->addWidget(icp);
    vbox->addWidget(icp_svm);

    QFrame* line = new QFrame();
    line->setFrameShape(QFrame::HLine);
    line->setFrameShadow(QFrame::Sunken);
    vbox->addWidget(line,1);
    this->setWidget(contents);

    refHist = new QLabel;
    vbox->addWidget(refHist);
    connect(_viewer, SIGNAL(updateScRefHistogram()), this, SLOT(updateRefHistogram()));
    connect(_viewer, SIGNAL(selectionTypeChange(TYPE)), this, SLOT(changeTypeOfVisuDock(TYPE)));

    curHist = new QLabel;
    vbox->addWidget(curHist);
    distHist = new QLabel;
    vbox->addWidget(distHist);
    connect(_viewer, SIGNAL(updateScCurHistogram()), this, SLOT(updateCurHistogram()));

    QButtonGroup* textureBox = new QButtonGroup();

    QCheckBox* HSV = new QCheckBox();
    HSV->setText("hsv texture");
    QCheckBox* field_text = new QCheckBox();
    field_text->setText("field texture");

    HSV->setChecked(true);
    // caution here the id of each button should be the same of the enum number in VISU_MODE (glviewer.h)
    textureBox->addButton(HSV, 0);
    textureBox->addButton(field_text, 1);
    textureBox->setExclusive(true);
    vbox->addWidget(HSV);
    vbox->addWidget(field_text);


    connect(textureBox, SIGNAL(buttonClicked(int)),_viewer, SLOT(changeDebugTexture(int)));

    QHBoxLayout *hbox = new QHBoxLayout();
    vbox->addLayout(hbox);
}


void VisuDock::updateRefHistogram(){
    Trace(1, " DockUser : updateHistogram ");
     refHist->setPixmap(_viewer->getSCRefHistogram());
     refHist->setVisible(true);
}

void VisuDock::updateCurHistogram(){
    Trace(4, " DockUser : updateHistogram ");
    float dist;
    curHist->setPixmap(_viewer->getSCCurHistogram(dist));
    curHist->setVisible(true);
    distHist->setText("histograms distance: " + QString::number(dist));
}

void VisuDock::changeTypeOfVisuDock(TYPE t){
    switch (t){
    refHist->setVisible(false);
    distHist->setVisible(false);
    case PATCH:
        refHist->setVisible(true);
        distHist->setVisible(true);
        break;
    case PART:
        break;
    case COMP:
        refHist->setVisible(true);
        distHist->setVisible(true);
        break;
    default:
        break;
    }
}
